
/*=============================================================================
   beest, an HTML viewer.
   Copyright (C) 2000 Zachary Thayer Smith

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   The author is reachable by electronic mail at zeyr@excite.com.
=============================================================================*/

/*----------------------------------------------------------------------
 * Module name:    fonthash.c
 * Author name:    Zach Smith
 * Create date:    1 Oct 2000
 * Last change:    26 Nov 2000
 * Purpose:        Hashing of Font information; smart retrieval of Font
 *                 structures.
 *----------------------------------------------------------------------
 * Changes:
 *--------------------------------------------------------------------*/



#include <stdio.h>
#include <ctype.h>


#include "defs.h"
#include "main.h"
#include "font.h"
#include "fonthash.h"
#include "error.h"
#include "font.h"
#include "bdf.h"




static Font *hash[] = {
	NULL, NULL, NULL, NULL, NULL, NULL, 
	NULL, NULL, NULL, NULL, NULL, NULL, 
	NULL, NULL, NULL, NULL, NULL, NULL, 
	NULL, NULL, NULL, NULL, NULL, NULL, 
	NULL, NULL, NULL, NULL, NULL, NULL, 
	NULL, NULL, NULL, NULL, NULL, NULL, 
};



/* Routine to generate a hash index based on 
 * the font family name and font size.
 */
static int
get_hash_index (char *family)
{
	int index;

	CHECK_PARAM_NOT_NULL (family);

	index = tolower (*family);
	if (index < 'a' || index > 'z') {
		warning_handler ("font family name does not begin with letter. ignored...");
		return -1;
	}
	index -= 'a';

	return index;
}



void
fonthash_store (Font *f) {
	int index;

	CHECK_PARAM_NOT_NULL (f);

	index = get_hash_index (f->family);
	if (index < 0) return;
	f->next = hash[index];
	hash[index] = f;
}



/* Some fonts are equivalent to others, but the HTML coder may not know
 * that. If they do, they will code that into the HTML, e.g.
 * <font face=arial,helvetica> however here we attempt to do that
 * for them. We recognize the presence of certain fonts that are
 * compiled into the program as being alternatives to whatever they
 * are asking for.
 */
static
char *font_equiv (char *family)
{
	if (!family || *family==0) 
		return "Times";

	switch (tolower(*family)) {
	case 'h':
		if (my_case_insensitive_strcmp (family, "HelveticaNarrow"))
			return "Helvetica";
		if (my_case_insensitive_strcmp (family, "Helvetica Narrow"))
			return "Helvetica";
		break;
	case 'a':
		if (my_case_insensitive_strcmp (family, "Arial"))
			return "Helvetica";
		break;
	case 'p':
		if (my_case_insensitive_strcmp (family, "Palatino"))
			return "NewCentury";
		break;
	case 'n':
		if (my_case_insensitive_strcmp (family, "NewCenturySchoolbook"))
			return "New Century Schoolbook";
		if (my_case_insensitive_strcmp (family, "New Times Roman"))
			return "Times";
		if (my_case_insensitive_strcmp (family, "NewTimesRoman"))
			return "Times";
		break;
	case 't':
		if (my_case_insensitive_strcmp (family, "Times Roman"))
			return "Times";
		if (my_case_insensitive_strcmp (family, "TimesRoman"))
			return "Times";
		break;
	}
	return family;
}


int
fonthash_has_family (DisplayDevice *dpy, char *name) {
	int index;
	Font *f;

	if (!name) return FALSE;

	index = get_hash_index (name);
	if (index < 0) 
		return FALSE;

	f = hash[index];

	while (f)
	{
		if (my_case_insensitive_strcmp (name, f->family))
			return TRUE;
		f = f->next;
	}
	return FALSE;
}


/* Assuming that the caller knows that this family is
 * present in the hash, 
 */
Font*
fonthash_find_closest_size_match (char *family, FontSize size, unsigned short attr) 
{
	int index;
	Font *f;
	signed long distance = 100000;
	unsigned long best_size = 100000;
	Font *best_font = NULL;
	unsigned char b0,b1,s0,s1;

	CHECK_PARAM_NOT_NULL(family);

	if (debug_mode && logfile)
	  fprintf (logfile, "Entered fonthash_find_closest_size_match seeking family=%s size=%d attr=%u\n",family,(short)size,attr);

	b0 = attr & ATTR_BOLD ? 1:0;
	s0 = attr & ATTR_ITALIC ? 1:0;

	index = get_hash_index (family);
	if (index < 0) 
		return NULL;

	f = hash[index];

	while (f)
	{
		b1 = f->is_bold ? 1:0;
		s1 = f->is_slanted ? 1:0;

		/* Compare, making all the quick comparisons first,
		 * then finally the strings.
		 */
		if (b0==b1 && s0==s1 && 
		    my_case_insensitive_strcmp (family, f->family)) 
		{
			signed long d = size - f->size;
			if (d<0) d = -d;
			if (d < distance) 
			{
				distance = d;
				best_size = f->size;
				best_font = f;
			}
		}
		f = f->next;
	}
	return best_font;
}


/*-----------------------Font Size Logic--------------------------------*/

static FontSize standard_font_sizes[] = {

	/* These, I believe, are standard sizes */
	8,10,12,14,18,24,36,48,

	/* These I just made up, I don't know if they are standard... */
	64,72,80,96,108,120,160,192,208,

	/* This is required */
	INVALID_FONTSIZE,
};
#define TOTAL_STD_FONTSIZES (sizeof(standard_font_sizes)/sizeof(FontSize))


FontSize 
determine_bigger_standard_size (FontSize value) {
	int i=0;
	while (i<TOTAL_STD_FONTSIZES) {
		if (value <= standard_font_sizes [i]) 
			break;
		i++;
	}
	if (i>=(TOTAL_STD_FONTSIZES-1))
		value = standard_font_sizes [TOTAL_STD_FONTSIZES-1];
	else
		value = standard_font_sizes [i+1];

	return value;
}


FontSize 
determine_smaller_standard_size (FontSize value) {
	int i=0;
	while (i<TOTAL_STD_FONTSIZES) {
		if (value <= standard_font_sizes [i]) break;
		i++;
	}
	if (i<2)
		value = standard_font_sizes [0];
	else
		value = standard_font_sizes [i-1];

	return value;
}


void
fonthash_init ()
{
	Font *f;

/*----------------*/
	
	f=font_new(); getbdf_TimesRM8 (f); 
	fonthash_store(f);
	f=font_new(); getbdf_TimesRM10 (f); 
	fonthash_store(f);
	f=font_new(); getbdf_TimesRM12 (f); 
	fonthash_store(f);
	f=font_new(); getbdf_TimesRM14 (f); 
	fonthash_store(f);
	f=font_new(); getbdf_TimesRM18 (f); 
	fonthash_store(f);
	f=font_new(); getbdf_TimesRM24 (f); 
	fonthash_store(f);

	f=font_new(); getbdf_HelveticaRM8 (f); 
	fonthash_store(f);
	f=font_new(); getbdf_HelveticaRM10 (f); 
	fonthash_store(f);
	f=font_new(); getbdf_HelveticaRM12 (f); 
	fonthash_store(f);
	f=font_new(); getbdf_HelveticaRM14 (f); 
	fonthash_store(f);
	f=font_new(); getbdf_HelveticaRM18 (f); 
	fonthash_store(f);
	f=font_new(); getbdf_HelveticaRM24 (f); 
	fonthash_store(f);

	f=font_new(); getbdf_CourierRM8 (f); 
	fonthash_store(f);
	f=font_new(); getbdf_CourierRM10 (f); 
	fonthash_store(f);
	f=font_new(); getbdf_CourierRM12 (f); 
	fonthash_store(f);
	f=font_new(); getbdf_CourierRM14 (f); 
	fonthash_store(f);
	f=font_new(); getbdf_CourierRM18 (f); 
	fonthash_store(f);
	f=font_new(); getbdf_CourierRM24 (f); 
	fonthash_store(f);

	f=font_new(); getbdf_NewCenturySchoolbookRM8 (f); 
	fonthash_store(f);
	f=font_new(); getbdf_NewCenturySchoolbookRM10 (f); 
	fonthash_store(f);
	f=font_new(); getbdf_NewCenturySchoolbookRM12 (f); 
	fonthash_store(f);
	f=font_new(); getbdf_NewCenturySchoolbookRM14 (f); 
	fonthash_store(f);
	f=font_new(); getbdf_NewCenturySchoolbookRM18 (f); 
	fonthash_store(f);
	f=font_new(); getbdf_NewCenturySchoolbookRM24 (f); 
	fonthash_store(f);

}



Font *
fonthash_lookup (DisplayDevice *dpy, String family, FontSize size, int attr)
{
	Font *f;

	/* If we don't have the family then just try to find something
	 * in the default font.
	 */
	if (!family) family=dpy->family;
	if (!fonthash_has_family (dpy, family)) {
		family = font_equiv (family);
		if (!fonthash_has_family (dpy, family)) {
			family = dpy->family;
		}
	}

	/* Finding the font is a process of elimination.
	 */
	f = fonthash_find_closest_size_match (family,size,attr);
	if (f) {
		return f;
	}

	/* Try an equivalent 
	 */
	family = font_equiv (family);

	/* Finding the font is a process of elimination.
	 */
	f = fonthash_find_closest_size_match (family,size,attr);
	if (f) {
		return f;
	}

	/* If not a perfect match, abandon the attributes
	 */
	if (attr) {
		attr = 0;
		f = fonthash_find_closest_size_match (family,size,attr);
		if (f) {
			return f;
		}
	}

	return NULL;
}

