
/*=============================================================================
   beest, an HTML viewer.
   Copyright (C) 2000 Zachary Thayer Smith

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   The author is reachable by electronic mail at zeyr@excite.com.
=============================================================================*/

/*----------------------------------------------------------------------
 * Module name:    style.c
 * Author name:    Zach Smith
 * Create date:    1 Oct 2000
 * Last change:    26 Nov 2000
 * Purpose:        Style class.
 *----------------------------------------------------------------------
 * Changes:
 *--------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>

#include "defs.h"
#include "main.h"
#include "malloc.h"
#include "error.h"
#include "style.h"
#include "group.h"
#include "font.h"
#include "fonthash.h"
#include "string.h"


void
style_reset (Style *sl) 
{
	CHECK_PARAM_NOT_NULL(sl);
	memset ((void*)sl, 0, sizeof(Style));
	sl->size = INVALID_FONTSIZE;
	sl->fg = INVALID_COLOR;
	sl->bg = INVALID_COLOR;
}


Style*
style_new() {
	Style *sl;

	sl = (Style*) my_malloc (sizeof (Style));
	if (!sl) error_handler ("unable to allocate Style object");
	style_reset (sl);
	return sl;
}


/* Some routines like to put things in Styles which are bogus.
 * Font families may be unknown or missing, font size may be some 
 * unheard of number.
 */
void
style_fix_font (DisplayDevice *dpy, Style *sl)
{
	int save = sl->flags;

	CHECK_PARAM_NOT_NULL (sl);

	if (sl->fontdata) 
		return;

	sl->flags = 0;
	sl->fontdata = fonthash_lookup (dpy, sl->family, sl->size, sl->flags);

	/* Restore the attributes, in case we
	 * are actually simulating them.
	 */
	sl->flags = save;

	if (!sl->fontdata) {
		sl->family = my_strdup(dpy->family);
		sl->fontdata = fonthash_lookup (dpy, sl->family, sl->size, sl->flags);

		if (!sl->fontdata) {
			sl->size = dpy->fontsize;
			sl->fontdata = fonthash_lookup (dpy, sl->family, sl->size, sl->flags);

			if (!sl->fontdata) {
error_handler ("failure to get even basic font data");
			}
		}
	}
}


void
style_dump(FILE *f, Style *sl)
{
	int attr;

	if (!sl) { fprintf (f, "(none)"); return; }

	fprintf (f, "Style {");
	if (sl->name)
		fprintf (f," name=%s ", sl->name);
	else
		fprintf (f," (no name) ");

	attr = sl->flags;
	fprintf (f,"font=%s pointsize=%u ic=%u sa=%u %s %s %s %s %s %s %s ",
		sl->family ? ((char*) sl->family) : "(none)", 
		(unsigned short) sl->size,
		(unsigned short) sl->interchar_spacing,
		(unsigned short) sl->super_amount,
		attr & ATTR_BOLD ? "bold" : "",
		attr & ATTR_ITALIC ? "ital" : "",
		attr & ATTR_UNDERLINE ? "underline" : "",
		attr & ATTR_SUPER ? "superscript" : "",
		attr & ATTR_SUB ? "subscript" : "",
		attr & ATTR_PLAIN ? "plain" : "",
		attr & ATTR_STRIKE ? "strike-through" : ""
		);
	if (!sl->fontdata) {
		fprintf (f,"(no font data)");
	} else {
		font_dump (f, sl->fontdata);
		if (!sl->fontdata->bytes) {
			fprintf (f,"(no font pointers!)");
		}
	}
	fprintf (f,"}\n");
}


void
style_set_font (Style *sl, Font *f) 
{
	CHECK_PARAM_NOT_NULL(sl);

	if (!f) {
		sl->fontdata = NULL;
		if (sl->family) my_free (sl->family);
		sl->family = NULL;
		sl->size = INVALID_FONTSIZE;
	} else {
		sl->fontdata = f;
		if (sl->family) my_free (sl->family);
		sl->family = my_strdup (f->family);
		sl->size = f->size;

		if (sl->size != sl->fontdata->size ||
		    !my_case_insensitive_strcmp (sl->family, sl->fontdata->family))
			warning_handler ("Style font info is different than its Font's info");
	}

#if 0
	if (!sl->family && dpy->family) 
		sl->family = my_strdup (dpy->family);
	if (sl->size==INVALID_FONTSIZE && dpy->fontsize!=INVALID_FONTSIZE) 
		sl->size = dpy->fontsize;

	/* Need to request that the display module
	 * provide font metrics.
	 */
	sl->fontdata = dpy->get_font_metrics (sl->family, sl->size, sl->flags);

	if (!sl->fontdata) {
		my_free (sl->family);
		sl->family = my_strdup ("Times");

		sl->fontdata = dpy->get_font_metrics (sl->family, sl->size, sl->flags);

		if (!sl->fontdata) 
			error_handler ("unable to get font metrics");
	}
#endif
}


void
style_copy (Style *dst, Style *src) {
	unsigned char *d;
	unsigned char *s;

	CHECK_PARAM_NOT_NULL(dst);
	CHECK_PARAM_NOT_NULL(src);

	d = (unsigned char*) dst;
	s = (unsigned char*) src;

	memcpy (d,s,sizeof(Style));
	if (src->family)
		dst->family = my_strdup (src->family);
	if (src->name)
		dst->name = my_strdup (src->name);

	dst->next = NULL;
}


int
style_interpret_string (Style *sl, String str)
{
	/* for later addition of CSS plus inline styles */
	return FALSE;
}


#if 0
Style *
style_dup (Style *src) 
{
	Style *dst;

	CHECK_PARAM_NOT_NULL(src);

	dst = style_new();
	style_copy (dst, src);

	return dst;
}
#endif

void 
style_free(Style *sl) {
	CHECK_PARAM_NOT_NULL (sl);
	if (sl->name)
		my_free ((char*) sl->name);
	if (sl->family)
		my_free ((char*) sl->family);

	my_free ((char*) sl);
}



